/* $Id: virge.c,v 1.11 1999/10/06 11:24:08 jos Exp $
******************************************************************************

   S3 ViRGE acceleration support

   Copyright (C) 1999    Jos Hulzink	[josh@stack.nl]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#define	MAINTAINER	"Jos Hulzink, [josh@stack.nl]"
#define	DRIVER_VERSION	"0.01"
#define DEBUG_LEVEL 0

#include <kgi/module.h>
#include <kgi/kgi_commands.h>
#include <kgi/types.h>

#include <linux/fb.h>

#include <S3/virge.h>

#include "suggest.inc"

#define GET_CHIPSET_PRIVATE struct chipset_private *private=KGIM_PRIVATE(dpy,chipset)
#define MEM_ACCEL (((struct chipset_private*)(KGIM_PRIVATE(dpy,chipset)))->mem_accel)

#define VIRGEACCELOUT(d,a)	mem_out32(d,MEM_ACCEL+a)
#define VIRGEACCELIN(a)		mem_in32(MEM_ACCEL+a)

/* The Accellerator must be idle before any mem_in () in accelleration space */
#define VIRGEWAITACCELIDLE() 	while(!(VIRGEACCELIN(0x8504)& 0x2000)) schedule()
/* When writing to the engine, for each mem_out (), 1 slot must be free */
#define VIRGEWAITFIFOEMPTY(n)   while(((VIRGEACCELIN(0x8504) >> 8)& 0x1f) < n) schedule()

#define VIRGETO2D(d)	if (accelmode != MODE2D) ViRGE_2D_setup (d)
#define VIRGETO3D(d)	if (accelmode != MODE3D) ViRGE_3D_setup (d)

static kgiu32 csm; /* command screen mode */
static kgiu8 bpp; /* bits per pixel */
static enum {MODE2D,MODE3D} accelmode;
static kgiu32 base,stride;

static struct {
  kgiu32 base,size;
} zbuffer_region, texture_region;

inline static void ViRGE_2D_setup (struct kgi_display *dpy) {
  accelmode = MODE2D;
  VIRGEWAITFIFOEMPTY(3);
  VIRGEACCELOUT (base, 0xA5D4); /* SRC_BASE */
  VIRGEACCELOUT (base, 0xA5D8); /* DEST_BASE */
  VIRGEACCELOUT (stride * 0x10001, 0xA5E4); /* SRC/DEST_STRIDE */
  
  /* All off-screen buffers must be invalidated here ! */ 
  dpy->fb.zbuf=NULL;
  zbuffer_region.base=zbuffer_region.size=
  texture_region.base=texture_region.size=0;
}

inline static void ViRGE_3D_setup (struct kgi_display *dpy) {
  GET_CHIPSET_PRIVATE;

  if (((dpy->mode.fb_size-dpy->mode.fsize-1024)>>1)>(dpy->mode.request.virt.x*dpy->mode.request.virt.y)) {
    accelmode = MODE3D;
    zbuffer_region.base=(dpy->mode.fsize+7)&(~7);
    zbuffer_region.size=(dpy->mode.request.virt.x*dpy->mode.request.virt.y*2);
    texture_region.base=zbuffer_region.base+zbuffer_region.size;
    texture_region.size=dpy->mode.fb_size-texture_region.base-1024;
    DEBUG1 ("ViRGE: Z Buffer space at offset: %d, size: %d",zbuffer_region.base,zbuffer_region.size);
    DEBUG1 ("ViRGE: Texture space at offset: %d, size: %d",texture_region.base, texture_region.size);
  
    VIRGEWAITFIFOEMPTY(2);
    VIRGEACCELOUT (zbuffer_region.base, 0xB0D4);
    VIRGEACCELOUT (dpy->mode.request.virt.x << 1, 0xB0E8);
   
    dpy->fb.zbuf=private->mem_law_region.baseptr+zbuffer_region.base;
  } else {
    accelmode = MODE2D; /* No 3D accelleration possible in this mode */
    zbuffer_region.base=zbuffer_region.size=
    texture_region.base=texture_region.size=0;
    dpy->fb.zbuf=NULL;
  }
}

inline static void ViRGE_set_colors (int fg, int bg, struct kgi_display *dpy) {
  VIRGEWAITFIFOEMPTY(2);
  VIRGEACCELOUT (fg, 0xA4F4);
  VIRGEACCELOUT (bg, 0xA4F8);
}

inline static int ViRGE_2D_draw_box (struct kgi_box *b,struct kgi_display *dpy) {
  DEBUG1("ViRGE: Accellerated DrawBox");
  VIRGETO2D (dpy);
  VIRGEWAITFIFOEMPTY(3);
  VIRGEACCELOUT (0x17E00121 | csm, 0xA500);
  VIRGEACCELOUT (b->h + ((b->w - 1) << 16), 0xA504);
  VIRGEACCELOUT (b->y + ((b->x) << 16), 0xA50C);
  return -EOK;
}

inline static int ViRGE_2D_copy_box (struct kgi_cbox *b,struct kgi_display *dpy) {
  DEBUG1("ViRGE: Accellerated CopyBox");
  VIRGETO2D (dpy);
  VIRGEWAITFIFOEMPTY(4);
  VIRGEACCELOUT (0x01980021 | csm | ((b->x >= b->x2) ? 0x02000000 : 0) | 
            ((b->y >= b->y2) ? 0x04000000 : 0), 0xA500);   
  VIRGEACCELOUT (b->h + ((b->w - 1) << 16), 0xA504);
  VIRGEACCELOUT ((((b->x >= b->x2) ? b->x : b->x + b->w - 1) << 16) + 
              ((b->y >= b->y2) ? b->y : b->y + b->h - 1), 0xA508);
  VIRGEACCELOUT ((((b->x >= b->x2) ? b->x2 : b->x2 + b->w - 1) << 16) +
              ((b->y >= b->y2) ? b->y2 : b->y2 + b->h - 1), 0xA50C);
  return -EOK;
}

inline static int ViRGE_2D_draw_line (struct kgi_line *l, struct kgi_display *dpy) {
  kgis32 xdelta;
  int tx,ty,xmayor;
  kgiu32 xstart; 

  DEBUG1("ViRGE: Accellerated DrawLine");
  DEBUG2("ViRGE: Line:(%d,%d) - (%d,%d)", l->x, l->y, l->x2, l->y2);

  if (l->y < l->y2) {
    tx = l->x; l->x = l->x2; l->x2 = tx;
    ty = l->y; l->y = l->y2; l->y2 = ty;
  } 
  
  xmayor = abs(l->x - l->x2) > (l->y - l->y2); 

  if (l->y == l->y2) 
    xdelta = 0;
  else
    xdelta = ((l->x - l->x2) << 20) / (l->y2 - l->y);
  
  if (xmayor)
    if (xdelta >= 0) xstart = (l->x << 20) + xdelta / 2;
    else xstart = (l->x << 20) + xdelta / 2 + ((1 << 20) - 1);
  else xstart = (l->x << 20);  
  
  VIRGETO2D(dpy);
  VIRGEWAITFIFOEMPTY(6);
  VIRGEACCELOUT (0x19E00121 | csm, 0xA900);
  VIRGEACCELOUT (l->x2 + (l->x << 16), 0xA96C);
  VIRGEACCELOUT (xdelta,  0xA970);
  VIRGEACCELOUT (xstart,  0xA974);
  VIRGEACCELOUT (l->y,  0xA978);
  VIRGEACCELOUT (l->y - l->y2 + ((l->x > l->x2) ? 1: 0x80000001), 0xA97C);
  return -EOK;
}

inline static int ViRGE_2D_draw_hline (struct kgi_hline *l, struct kgi_display *dpy) {
  DEBUG1("ViRGE: HLine:(%d-%d,%d)",l->x, l->x+l->w-1, l->y);
  VIRGETO2D(dpy);
  VIRGEWAITFIFOEMPTY(3);
  VIRGEACCELOUT (0x17E00121 | csm, 0xA500);   
  VIRGEACCELOUT (1 + ((l->w - 1) << 16), 0xA504);
  VIRGEACCELOUT (l->y + ((l->x) << 16), 0xA50C);
  return -EOK;
}

inline static int ViRGE_2D_draw_vline (struct kgi_hline *l, struct kgi_display *dpy) {
  DEBUG1("ViRGE: VLine:(%d,%d-%d) ",l->x, l->y, l->y+ l->w-1);
  VIRGETO2D(dpy);
  VIRGEWAITFIFOEMPTY(3);
  VIRGEACCELOUT (0x17E00121 | csm, 0xA500);   
  VIRGEACCELOUT (l->w, 0xA504);
  VIRGEACCELOUT (l->y + ((l->x) << 16), 0xA50C);
  return -EOK;
}

inline static int ViRGE_2D_fill_screen (kgi_modereq mode, struct kgi_display *dpy) {
  DEBUG1("ViRGE: Accellerated FillScreen");	  	
  VIRGETO2D(dpy);
  VIRGEWAITFIFOEMPTY(3);
  VIRGEACCELOUT (0x17E00121 | csm, 0xA500);   
  VIRGEACCELOUT (mode.visible.y + (mode.visible.x << 16), 0xA504);
  VIRGEACCELOUT (0, 0xA50C);
  return -EOK;
}

/*
**	KGI module interface
*/

int kgim_accel_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE("ViRGE: accelleration driver loaded");
	return EOK;
}

void kgim_accel_done(struct kgi_display *dpy)
{
	NOTICE("ViRGE: accelleration driver removed");
}

int kgim_accel_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
 			  enum kgi_tm_cmd *cmd)
{
	return EOK;
}

void kgim_accel_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
 	switch (mode->request.graphtype) {
 	  case KGIGT_8BIT: 
 	    csm=0;
 	    bpp=8;
 	    break;
 	  case KGIGT_15BIT:
 	  case KGIGT_16BIT:
 	    csm=0x4;
 	    bpp=16;
 	    break;
 	  case KGIGT_24BIT:
 	    csm=0x8;
 	    bpp=24;
 	    break;
 	  case KGIGT_32BIT:
 	    csm=0x8;
 	    bpp=32; /* Not supported ??? */
 	    break;
 	  default:
 	    csm=0;
 	    bpp=8;
 	    break;   
 	} 

	base = 0;
	stride = ((bpp*mode->request.virt.x) / 8);	

  	ViRGE_2D_setup (dpy); 
}


int kgim_accel_command(struct kgi_graphic *gr, kgi_uint cmd, kgi_uint arg)
{ 
	kgi_suggest *sugg;
	const char *name;

	switch (cmd) 
	{
	case ACCEL_DRAWHLINE:
	  ViRGE_set_colors (gr->gc->fg_color, gr->gc->bg_color, gr->dev.dpy);
	  return ViRGE_2D_draw_hline ((struct kgi_hline *)(gr->io_buf),gr->dev.dpy);
	case ACCEL_PUTHLINE:
	case ACCEL_GETHLINE:
	  return -E(ACCEL, NOSUP_ALWAYS_MMAP);
	case ACCEL_DRAWLINE:
	  ViRGE_set_colors (gr->gc->fg_color, gr->gc->bg_color, gr->dev.dpy);
	  return ViRGE_2D_draw_line ((struct kgi_line *)(gr->io_buf),gr->dev.dpy);
	case ACCEL_DRAWVLINE:
	  ViRGE_set_colors (gr->gc->fg_color, gr->gc->bg_color, gr->dev.dpy);
	  return ViRGE_2D_draw_vline ((struct kgi_hline *)(gr->io_buf),gr->dev.dpy);
	case ACCEL_PUTVLINE:
	case ACCEL_GETVLINE:
	  return -E(ACCEL, NOSUP_ALWAYS_MMAP);
	case ACCEL_DRAWBOX:
	  ViRGE_set_colors (gr->gc->fg_color, gr->gc->bg_color, gr->dev.dpy);
	  return ViRGE_2D_draw_box ((struct kgi_box *)(gr->io_buf),gr->dev.dpy);
	case ACCEL_GETBOX:
	case ACCEL_PUTBOX:
	  return -E(ACCEL, NOSUP_ALWAYS_MMAP);
	case ACCEL_FILLSCREEN:
	  ViRGE_set_colors (gr->gc->fg_color, gr->gc->bg_color, gr->dev.dpy);
	  return ViRGE_2D_fill_screen (gr->dev.mode, gr->dev.dpy);
	case ACCEL_COPYBOX:
	  return ViRGE_2D_copy_box ((struct kgi_cbox *)(gr->io_buf),gr->dev.dpy);
	case ACCEL_GETSUGGEST:
		sugg=(kgi_suggest *)(gr->io_buf);
		gr->io_res=gr->io_buf;

		if (sugg->id<0 || sugg->id > 1)
			return -E(ACCEL, BOGUS);

		if (sugg->id==1) { /* Generic RAMDAC support */

			name = sugg_default[gr->dev.mode.graphtype].text[1];
			strcpy(sugg->name, name);
			sugg->args[0] = 0;
			sugg->id = 0;
			return EOK;
		}
		name=sugg_default[gr->dev.mode.graphtype].text[0];
		strcpy(sugg->name,name);
		sugg->args[0] = 0;
		sugg->id = 1;	/* Next is RAMDAC.. */
		return EOK;

	default:
		break;
	}
			
	return -E(ACCEL, NOSUP_ALWAYS_MMAP);
}
